#pragma TextEncoding	= "UTF-8"
#pragma rtGlobals		= 3	
#pragma IgorVersion		= 6.36
#pragma ModuleName		= DifferenceGenerator
#pragma version			= 4.4
#pragma DefaultTab		= {3,20,4}																// Set default tab width in Igor Pro 9 and later

// --------------------- Project Updater header ----------------------
// If you're using Igor Pro 8 or later and have Project Updater
// installed, this package can check periodically for new releases.
// https://www.wavemetrics.com/project/Updater
static Constant kProjectID = 21539 																// the project node on IgorExchange
static StrConstant ksShortTitle = "Difference Generator"										// the project short title on IgorExchange

//________________________________________________________________________________________________
//	Written by Stephan Thuermer - https://www.wavemetrics.com/user/chozo
//	Shift, scale, offset and compare data against reference data
//	
// 2021-02-04 - ver. 4.20:	Initial public release.
// 2021-02-05 - ver. 4.30:	Added multiplication and division modes.
// 2021-03-07 - ver. 4.31:	Forgot to remove debug print command.
// 2021-03-17 - ver. 4.32:	Gauss convolution was not working well if data included NaNs.
//							Now the smoothing value is also displayed in the legend.
// 2021-03-19 - ver. 4.33:	Fixed an error message when loading specific old data sets.
// 2021-08-14 - ver. 4.40:	Improved edge behavior of the Gaussian smoothing (data edge are less likely to bulge).
//							Moved all controls into an external panel.
//							Added controls to directly set absolute X,Y, and scale values for changes.
//							Now Background Pinning honors smoothing as well.
//________________________________________________________________________________________________

static Constant	kSaveAux = 0																	// 1 = additionally generated data (currently the smoothed difference) will be saved
static Constant	kAuxMode = 0																	// what is displayed as auxiliary data: 0 = smoothed difference; 1 = integrated difference
static StrConstant kModEnd	= "_mod"															// label for modified data (max. 4 characters for Igor version < 7)
static StrConstant kOutEnd	= "_dif;_sum;_mul;_div;"											// label for difference data
static StrConstant kAuxEnd	= "_smt;_int;"														// label for additional data => same order as kAuxMode
static StrConstant kDiffGenDir	= "Difference_Generator"										// name for the working folder inside Packages
static StrConstant kDiffTitle	= "Spectra Arithmetic Tool (version 4.4 - 07/2021): Modifying "	// title used for the graph

// the following constants define the string keys for saving and loading parameters from wave notes
static StrConstant kScaleKey	= "Mod Scale"
static StrConstant kShiftKey	= "Mod Shift"
static StrConstant kOffsetKey	= "Mod Offset"
static StrConstant kModSmthKey	= "Mod Smooth"
static StrConstant kRefSmthKey	= "Ref Smooth"
static StrConstant kXCsrSetKey	= "Cursor Set"
static StrConstant kSetModeKey	= "Data difference;Data summed;Data multiplied;Data divided;"	// save key for all modes. The first one is not used for reloading, since this is the default mode.
static StrConstant kModeInvKey	= "Result inverted"
static StrConstant kPinBaseKey	= "Baseline scaled"
static StrConstant kRefNameKey	= "Reference name used"

Menu "Spectra Tools"
	"Difference between Spectra ...",/Q, CreateBrowser; DiffGen_LoadWrapper()
End

Function DiffGen_LoadWrapper()																	// calls DiffGen with the browser selection
	Wave/Z work = $GetBrowserSelection(0)	// first selection
	Wave/Z ref = $GetBrowserSelection(1)	// second selection
	if (WaveExists(work) == 0)
		return -1
	endif
	if (WaveExists(ref))
		SpectraDifferenceGenerator(work,refwave=ref)
	else
		SpectraDifferenceGenerator(work)
	endif
	return 0
End

//################################################################################################

Function SpectraDifferenceGenerator(inwave,[refwave])
	Wave inwave
	Wave refwave
	
	if (!WaveExists(inwave))
		return -1
	endif
	if (WaveType(inwave,1) != 1 || (WaveExists(refwave) && WaveType(refwave,1) != 1))
		Abort "This is meant for numeric waves!"
	endif
	if (WaveDims(inwave) > 1 || (WaveExists(refwave) && WaveDims(refwave) > 1))
		Abort "This is meant for 1D waves!"
	endif
#if IgorVersion() < 7.00
	if (strlen(NameOfWave(inwave)) > 27)														// abort, if the wave name (plus the 4 character addition later) may exceed the 31 char limit (Igor 6)
		Abort "Wave name is too long ... 27 characters at maximum."
	endif
#endif

	if (WinType("DiffGenGraphMainWindow"))														// make sure that there is only one panel open
		DoWindow/F DiffGenGraphMainWindow
		DoAlert/T="Open Difference Generator window found." 1, "There is another instance open. Close the old one and load the new data? Any changes will not be saved!"
		if (V_flag != 1)
			return 0
		endif
		KillWindow $getGraph()
	endif
	//+++++++++++++++++++++++++++++++++ generate working variables ++++++++++++++++++++++++++++++
	NewDataFolder/O	root:Packages
	NewDataFolder/O	root:Packages:$(kDiffGenDir)
	DFREF storage = root:Packages:$(kDiffGenDir)
	String/G		storage:locOriginal		= GetWavesDataFolder(inwave,2)
	String/G		storage:locReference	= ""
	Make/D/O		storage:Modifiers		= {0,0,1}											// contains temporary parameters: [0] = horizontal shift, [1] = vertical shift, [2] = multiplicator
	//+++++++++++++++++++++++++++++++++++++ build GUI window +++++++++++++++++++++++++++++++++++++
	String scnNum	= StringByKey("NSCREENS", IgorInfo(0))										// number of screens (assumed to have the same resolution)
	String scnInfo	= StringByKey("SCREEN"+scnNum,IgorInfo(0),":")
	Variable TFTx	= str2num(StringFromList(3,scnInfo,",")) - str2num(StringByKey("RECT",scnInfo,"=",","))
	Variable TFTy	= str2num(StringFromList(4,scnInfo,",")) - str2num(StringFromList(2,scnInfo,","))
	Variable pix = 72/ScreenResolution
	TFTx = TFTx > 0 && numtype(TFTx) == 0 ? TFTx : 1024											// fallback screen res
	TFTy = TFTy > 0 && numtype(TFTy) == 0 ? TFTy : 768
	Variable WinTop = 70, WinLeft = 200
	Variable WinBottom	= TFTy - 150
	Variable WinRight	= TFTx - 350
	
	Display/K=2/W=(WinLeft*pix, WinTop*pix, WinRight*pix, WinBottom*pix)/N=DiffGenMainGraph as kDiffTitle + NameOfWave(inwave)
	SetWindow DiffGenMainGraph hook(DiffWindow)=DiffGen_ProcessWindowEvent						// hook function processes key events
	//++++++++++++++++++++++++++++++ external panel with the controls +++++++++++++++++++++++++++++
	NewPanel/HOST=DiffGenMainGraph/EXT=0/K=2/N=DiffGenGraphMainWindow/W=(0, 0, 250, 448) as "Spectra Modification"
	SetWindow DiffGenMainGraph#DiffGenGraphMainWindow hook(DiffWindow)=DiffGen_ProcessWindowEvent	// keeps track of window changes here as well (only button presses)
	
	PopUpMenu ArithMode		,pos={10,8}		,size={175,23}	,title="Mode:"		,fstyle=1		,help={"Selects the current operation mode: A-B difference (- key), A+B addition (+ key), A*B multiplication (* key), A/B division (/ key)."}
	CheckBox InvertMode		,pos={195,10}	,size={50,23}	,title="\f04I\f00nvert"				,help={"Inverts the difference spectrum. (key: 'i')"}
	
	PopUpMenu OrigSelect	,pos={10,35}	,size={200,23}	,title="Modifying:"					,help={"Select the wave which will be edited."}
	PopUpMenu RefSelect		,pos={10,60}	,size={200,23}	,title="Reference:"					,help={"Select the reference wave to compare against."}
	Button bProgSwitch		,pos={220,35}	,size={22,44}	,title="]"			,fsize=20		,help={"Switches editing and reference waves."}
	
	GroupBox ConvLabel		,pos={8,90}		,size={235,45}	,title="Smoothing (Gauss convolution):"	,frame=0
	SetVariable WaveConv	,pos={20,110}	,size={105,23}	,title="Mod.:"						,help={"Convolves the modified wave with a Gaussian."}
	SetVariable RefWaveConv	,pos={130,110}	,size={105,23}	,title="Ref.:"						,help={"Convolves the reference wave with a Gaussian. Modification will not be saved."}

	GroupBox RelChangeLabel	,pos={8,140}	,size={235,133}	,title="Relative changes:"				,frame=0
	SetVariable Shift		,pos={15,162}	,size={130,23}	,title="delta X:"					,help={"The delta value by which the horizontal position will be shifted every step."}
	SetVariable Offset		,pos={15,191}	,size={130,23}	,title="delta Y:"					,help={"The delta value by which the vertical position will be shifted every step."}
	Button bUp				,pos={185,160}	,size={22,22}	,title="^"							,help={"Add a constant value to the data every step. Hold 'shift' ('ctrl/command') to accelerate (decelerate) the movement. Very fine movement with both ctrl/command + shift. (key: up arrow)"}
	Button bDown			,pos={185,188}	,size={22,22}	,title="v"							,help={"Remove a constant value from the data every step. Hold 'shift' ('ctrl/command') to accelerate (decelerate) the movement. Very fine movement with both ctrl/command + shift. (key: down arrow)"}
	Button bLeft			,pos={155,188}	,size={22,22}	,title="<"							,help={"Shift the spectrum to lower x values. Hold 'shift' ('ctrl/command') to accelerate (decelerate) the movement. Very fine movement with both ctrl/command + shift. (key: left arrow)"}
	Button bRight			,pos={213,188}	,size={22,22}	,title=">"							,help={"Shift the spectrum to higher x values. Hold 'shift' ('ctrl/command') to accelerate (decelerate) the movement. Very fine movement with both ctrl/command + shift. (key: right arrow)"}
	
	SetVariable Multiplier	,pos={15,220}	,size={130,23}	,title="Scale by:"					,help={"The intensity will be multiplied by this value every step."}			
	CheckBox PinBase		,pos={30,247}	,size={80,23}	,title="\f04P\f00in down Baseline:"	,help={"The baseline (wave minimum) stays fixed and will not be scaled. The fix point can be selected by setting cursor A or B on the reference wave (WaveR). (key: 'p')"}
	Button bStretch			,pos={155,220}	,size={80,20}	,title="Stretch"					,help={"Multiply intensity by the set percentage. Hold 'shift' ('ctrl/command') to accelerate (decelerate) the stretching. Very fine movement with both ctrl/command + shift. (key: page up)"}
	Button bShrink			,pos={155,245}	,size={80,20}	,title="Shrink"						,help={"Downsize intensity by the set percentage. Hold 'shift' ('ctrl/command') to accelerate (decelerate) the shrinking. Very fine movement with both ctrl/command + shift. (key: page down)"}
	
	GroupBox AbsChangeLabel	,pos={8,277}	,size={235,103}	,title="Absolute changes:"				,frame=0
	SetVariable AbsXshift	,pos={15,298}	,size={130,23}	,title="X:"							,help={"Set the absolute horizontal shift."}
	SetVariable AbsYshift	,pos={15,326}	,size={130,23}	,title="Y:"							,help={"Set the absolute vertical shift."}
	SetVariable AbsScale	,pos={15,354}	,size={130,23}	,title="Scale:"						,help={"Set the absolute scale."}

	Button bUndo			,pos={155,298}	,size={80,20}	,title="\f05R\f01eset:\f00 All"		,help={"Reset all modifications. Hold 'ctrl/command' to additionally unload the reference wave. (key: 'r')"}
	Button bUndoShift		,pos={155,325}	,size={35,20}	,title="X"							,help={"Reset horizontal shift."}
	Button bUndoOffset		,pos={200,325}	,size={35,20}	,title="Y"							,help={"Reset vertical shift."}
	Button bUndoScale		,pos={155,352}	,size={80,20}	,title="Scale"						,help={"Reset scaling."}
	
	Button bProgUndoQuit	,pos={10,390}	,size={60,44}	,title="Discard\r& Quit"			,help={"Discard changes and quit."}
	Button bProgSaveQuit	,pos={80,390}	,size={55,44}	,title="\f01Save\r& Quit"			,help={"Save changes and quit. Hold 'ctrl/command' key to just save the result without quitting."}	,valueColor=(65535,0,0)
	CheckBox FinishGraph	,pos={145,392}	,size={95,20}	,title="Display Result"				,help={"Display results graph at the end."}
	CheckBox AltSaveOption	,pos={145,418}	,size={95,20}	,title="Use Ref. Name"				,help={"Saves the difference wave with the name of the reference wave instead of the modified wave."}
	
	if (CmpStr(IgorInfo(2), "Macintosh") == 0)
		ModifyControlList ControlNameList("DiffGenMainGraph#DiffGenGraphMainWindow",";","*") fsize=10
	endif
	
	ModifyControlList ControlNameList("DiffGenMainGraph#DiffGenGraphMainWindow", ";", "b*")	,proc=DiffGen_ButtonControl		// define the procedure for modify buttons
	PopUpMenu OrigSelect	,bodywidth=140	,value=DiffGen_CreateWaveList("Orig")			,proc=DiffGen_PopupWaveSelect
	PopUpMenu RefSelect		,bodywidth=140	,value=DiffGen_CreateWaveList("Ref")			,proc=DiffGen_PopupWaveSelect
	PopUpMenu ArithMode		,bodywidth=115	,value="Subtract (A-B);Add (A+B);Multiply (A*B);Divide (A/B);"	,proc=DiffGen_ModeSelect
	
	SetVariable Offset		,bodyWidth=80	// no limits and values for these controls (will be set in DiffGen_LoadWavesAndSettings())
	SetVariable Shift		,bodyWidth=80
	SetVariable Multiplier	,bodyWidth=80	,value=_NUM:1					,limits={0.01,inf,0.5}		,format="%g %"
	SetVariable AbsXshift	,bodyWidth=110	,value=storage:Modifiers[0]		,proc=DiffGen_WaveConvolve						// for the Abs controls this only updates the calculation
	SetVariable AbsYshift	,bodyWidth=110	,value=storage:Modifiers[1]		,proc=DiffGen_WaveConvolve	,format="%g"
	SetVariable AbsScale	,bodyWidth=90	,value=storage:Modifiers[2]		,proc=DiffGen_WaveConvolve	,format="%g x"
	SetVariable WaveConv	,bodyWidth=75									,proc=DiffGen_WaveConvolve
	SetVariable RefWaveConv	,bodyWidth=75	,value=_NUM:0					,proc=DiffGen_WaveConvolve
	CheckBox InvertMode														,proc=DiffGen_InvertMode
	CheckBox PinBase		,side=1			,value=1						,proc=DiffGen_PinBaseMode
	CheckBox FinishGraph					,value=1
	//++++++++++++++++++++++++ add the traces after the graph was built ++++++++++++++++++++++++++++	
	DoWindow/F DiffGenMainGraph
	String recreate = DiffGen_LoadWavesAndSettings()											// create duplicates of the original waves and reload parameters if available (needs to be executed AFTER window controls were built)
	Wave/SDFR=storage WaveO, WaveR, WaveD, WaveA												// Info: WaveO = wave to edit; WaveR = reference wave; WaveD = difference wave; WaveA = auxiliary wave
	
	AppendToGraph WaveO
	AppendToGraph/C=(0,0,0)/L=Ldiff WaveD
	AppendToGraph/C=(0,40000,20000)/L=Ldiff WaveA
	ModifyGraph axisEnab(left)={0.32,1}, axisEnab(Ldiff)={0,0.28}, freePos(Ldiff)=0				// result is displayed at the bottom 30% of the graph
	ModifyGraph zero(Ldiff)=3																	// add a zero line
	ModifyGraph mirror=1, minor=1, standoff=0
	ModifyGraph ZisZ=1, notation(left)=1, gfSize=10
	SetAxis/A=2 Ldiff
	SetAxis/A=2 left
	Legend/C/N=GraphInfoBox/J/F=0/A=LT/X=2.00/Y=2.00 ""											// create trace legend
	Legend/C/N=ModificationInfoBox/J/F=0/A=RT/X=2.00/Y=2.00/B=1 ""								// create legend for modification info
	DoUpdate
	
	DiffGen_UpdateGraphLegends()																// create first legend entries
	
	WaveR = NaN
	if(WaveExists(refwave))																		// if reference wave is provided in function call
		DFREF waveDF = GetWavesDataFolderDFR(refwave)
		DFREF saveDF = GetDataFolderDFR()
		SetDataFolder waveDF
			PopUpMenu RefSelect ,mode=1 ,popvalue=NameOfWave(refwave)							// select reference wave
			DiffGen_WaveSelect("RefSelect",NameOfWave(refwave))									// load reference wave
		SetDataFolder saveDF
	else																						// instead sets the old reference wave if available
		DiffGen_ReloadRefWaveSettings(recreate)													// reloads previous reference wave and sets relevant controls
	endif
	
	DiffGen_CalculateDifference()
	return 0
End
//++++++++++++++++++++++++++++++++++++++++ number rounding +++++++++++++++++++++++++++++++++++++++
Static Function ChangePrecision(value, numSigDigits)											// round values to given number of digits
	variable value, numSigDigits
	string str
	sprintf str, "%.*g\r", numSigDigits, value
	return str2num(str)
End

//################################################################################################

static Function/S getPanel()
	String panelStr = "DiffGenMainGraph#DiffGenGraphMainWindow"
	if (WinType("DiffGenMainGraph") != 1)														// compatibility with older versions
		panelStr = "DiffGenGraphMainWindow"
	endif
	return panelStr
End

//################################################################################################

static Function/S getGraph()
	String graphStr = "DiffGenMainGraph"
	if (WinType(graphStr) != 1)																	// compatibility with older versions
		graphStr = "DiffGenGraphMainWindow"
	endif
	return graphStr
End

//################################################################################################

Function DiffGen_CalculateDifference([undoPinning])
	Variable undoPinning																		// if pin background was toggled
	Variable rev = ParamIsDefault(undoPinning) ? 0 : undoPinning
	
	DFREF stg = root:Packages:$(kDiffGenDir)
	Wave/SDFR=stg WaveO, WaveR, WaveD, WaveA, Modifiers											// Modifiers: [0] = horizontal shift, [1] = vertical shift, [2] = multiplicator
	SVar/SDFR=stg locOriginal, locReference
	Wave origW = $locOriginal

	ControlInfo/W=$getPanel() PinBase
	Variable fixMin = V_Value
	ControlInfo/W=$getPanel() WaveConv
	Variable wSmooth = V_Value
	ControlInfo/W=$getPanel() RefWaveConv
	Variable refSmooth = V_Value
	ControlInfo/W=$getPanel() InvertMode														// normal or inverted
	Variable invertDif = V_Value
	ControlInfo/W=$getPanel() ArithMode
	Variable difMode = V_value
	//++++++++++++++++++++++++++++++++++++ modify the wave +++++++++++++++++++++++++++++++++++++++
	Variable minPoint, minCsrX; 	DiffGen_FindBaseline(minPoint, minCsrX)						// use either the wave minimum or cursor location

	Modifiers[0] = numtype(Modifiers[0]) != 0 ? 0 : Modifiers[0]								// first, make sure these are proper numbers
	Modifiers[1] = numtype(Modifiers[1]) != 0 ? 0 : Modifiers[1]
	Modifiers[2] = numtype(Modifiers[2]) != 0 ? 1 : Modifiers[2]
	
	WaveO = (origW - minPoint*fixMin)*Modifiers[2]												// Y-scaling (with or without min. offset)
	SetScale/P x DimOffset(origW,0)+Modifiers[0], DimDelta(origW,0), WaveUnits(origW,0), WaveO	// X-shift
	Variable smthDrift = DiffGen_WaveConvolveGaussHelper(WaveO, wSmooth)						// smoothing (which can change the min wave value)
	Modifiers[1] += rev*(minPoint*(1 - Modifiers[2]) - smthDrift)								// if the pin state was changed: add or subtract the scaled amount to the offset (prevents 'snapping')
	WaveO += Modifiers[1] + minPoint*fixMin	- smthDrift*fixMin									// Y-shift (after smoothing)
	//+++++++++++++++++++++++++++ now calculate the difference +++++++++++++++++++++++++++++++++++
	Variable X1 = pnt2x(WaveO,0)
	Variable X2 = pnt2x(WaveO,numpnts(WaveO)-1)
	Variable mirrorO = sign(DimDelta(WaveO,0)) == 1												// mirror for negative delta values => reverses assignment for X1 and X2
	Variable mirrorR = sign(DimDelta(WaveR,0)) == 1
	SetScale/I x X1, X2, WaveUnits(WaveD,0), WaveD												// first reset everything
	WaveD = NaN; WaveA = NaN
	
	X1 = mirrorO ? pnt2x(WaveO,0) : pnt2x(WaveO,numpnts(WaveO)-1)								// check the left side
	X2 = mirrorR ? pnt2x(WaveR,0) : pnt2x(WaveR,numpnts(WaveR)-1)
	Variable Left = X1 > X2 ? X1 : X2															// the biggest value will be the left border
	X1 = mirrorO ? pnt2x(WaveO,numpnts(WaveO)-1) : pnt2x(WaveO,0)								// check the right side
	X2 = mirrorR ? pnt2x(WaveR,numpnts(WaveR)-1) : pnt2x(WaveR,0)
	Variable Right = X1 < X2 ? X1 : X2															// the smallest value will be the right border

	if (Right > Left)																			// look first if the waves are overlapping at all
		SetScale/I x Left, Right, WaveUnits(WaveD,0), WaveD, WaveA
		Switch (difMode)																		// here the difference will be generated
			case 1:
				WaveD  = WaveO(x) - WaveR(x)
			break
			case 2:
				WaveD  = WaveO(x) + WaveR(x)
			break
			case 3:
				WaveD  = WaveO(x) * WaveR(x)
			break
			case 4:
				WaveD  = WaveO(x) / WaveR(x)
			break
		EndSwitch
		WaveD *= invertDif ? -1 : 1
		WaveA = WaveD																			// Do something with the auxiliary wave
		Switch(kAuxMode)
			case 0:																				// smoothing
				Smooth round(DimSize(WaveD,0)*0.05), WaveA
				break
			case 1:
				Integrate WaveD /D=WaveA														// integration
				Variable Scale = abs(Wavemin(WaveD)) < abs(Wavemax(WaveD)) ? Wavemax(WaveD)/Wavemax(WaveA) : Wavemin(WaveD)/Wavemin(WaveA)
				ModifyGraph/Z/W=$getGraph() muloffset(WaveA)={0,Scale}
				break
		EndSwitch
	endif
	//+++++++++++++++++++++++ update the wave notes : save parameters ++++++++++++++++++++++++++++
	String NoteText = "Difference created from:"
	NoteText += "\rReference Wave="	+ locReference
	NoteText += "\rOriginal Wave="	+ locOriginal
	NoteText += "\rModification settings:"
	NoteText += "\r"+kShiftKey+"="	+num2str(Modifiers[0])
	NoteText += "\r"+kOffsetKey+"="	+num2str(Modifiers[1])
	NoteText += "\r"+kScaleKey+"="	+num2str(abs(Modifiers[2]))
	NoteText += "\r"+kModSmthKey+"="+num2str(wSmooth)
	
	if (numtype(minCsrX) == 0)
		NoteText += "\r"+kXCsrSetKey+"="+num2str(minCsrX)
	endif
	if (refSmooth > 0)
		NoteText += "\r"+kRefSmthKey+"="+num2str(refSmooth)
	endif
	if (difMode >= 0)
		NoteText += "\r"+StringFromList(difMode-1,kSetModeKey)
	endif
	if (invertDif)
		NoteText += "\r"+kModeInvKey
	endif
	if (fixMin == 0)
		NoteText += "\r"+kPinBaseKey
	endif
	
	String currNote = note(origW)
	Note/K WaveD, NoteText
	if (strlen(currNote))
		NoteText[0] = currNote + "\r\r"
	endif
	Note/K WaveO, NoteText
	//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
	DiffGen_UpdateGraphLegends()
	return 0
End
//+++++++++++++++++++++++++++++++++ which baseline value to use ++++++++++++++++++++++++++++++++++
Function DiffGen_FindBaseline(minPoint, minLoc)
	Variable &minPoint, &minLoc
	
	DFREF stg = root:Packages:$(kDiffGenDir)
	SVar/SDFR=stg locOriginal, locReference
	Wave origW = $locOriginal
	
	Variable csrAval = inf, csrBval = inf, csrAx = 0, csrBx = 0, useCsr = 0						// cursors may be used to set the baseline point
	if (StringMatch(CsrWave(A),"WaveR"))
		csrAval	= vcsr(A)
		csrAx	= xcsr(A)
		useCsr	= 1
	endif
	if (StringMatch(CsrWave(B),"WaveR"))
		csrBval	= vcsr(B)
		csrBx	= xcsr(B)
		useCsr	= 1
	endif
	if (useCsr)																					// save cursor location for recreation
		minLoc = csrAval < csrBval ? csrAx : csrBx
	else 
		minLoc = NaN
	endif
	
	minPoint = useCsr == 0 ? Wavemin(origW) : min(csrAval,csrBval)								// use either the wave minimum or cursor location
	return 0 
End
//################################################################################################
Function DiffGen_ProcessWindowEvent(s)
	STRUCT WMWinHookStruct &s

	Variable HookTakeover = 0
	Switch (s.EventCode)
		case 17:	// kill request
			HookTakeover = 1
			DFREF stg = root:Packages:$(kDiffGenDir)
			Wave/SDFR=stg WaveO
			SVar/SDFR=stg locOriginal
			Wave origwave  = $(locOriginal)
			Wave/Z modwave = $(ParseFilePath(1, locOriginal, ":", 1, 0) + PossiblyQuoteName(NameOfWave(origwave) + kModEnd))
			
			Variable changed = 0
			if (WaveExists(modwave))
				changed = !EqualWaves(WaveO, modwave, 5)
			else
				changed = !EqualWaves(WaveO, origwave, 5)
			endif
			
			if (changed)
				if (!Wintype("#AskQuitPanel"))													// this will create a little panel in case of changes
					Variable pix = 72/ScreenResolution
					Variable xCenter = (s.WinRect.right - s.WinRect.left)/2, yCenter = (s.WinRect.bottom - s.WinRect.top)/2
					NewPanel/N=AskQuitPanel/HOST=$s.WinName/W=((xCenter-105)*pix,(yCenter-38)*pix,(xCenter+105)*pix,(yCenter+38)*pix)
					DrawText 20,28,"Quiting: Save made changes?"
					Button bProgSaveQuit	,pos={20,33}	,size={50,30}	,title="\f01Save"	,proc=DiffGen_ButtonControl	,help={"Save changes and quit. Hold the 'alt/option' key to save the difference spectrum with the name of the reference wave instead of the editing wave."}
					Button bProgUndoQuit	,pos={80,33}	,size={50,30}	,title="Discard"	,proc=DiffGen_ButtonControl	,help={"Discard changes and quit."}
					Button bProgCancel		,pos={140,33}	,size={50,30}	,title="Cancel"		,proc=DiffGen_ButtonControl
					SetActiveSubwindow $s.WinName												// restore focus to main window
				endif
			else
				DiffGen_ProgramExecute("bProgUndoQuit")
			endif
			break
		case 11:	// keys
			HookTakeover = 1
			String Key = "", Cursors = ""
			Variable Ignore = 0																	// ignore = 1 will give Igor the control over this event
			Switch(s.keycode)																	// identify the pressed key
				case 11:
					Key = "bStretch"
					break
				case 12:
					Key = "bShrink"
					break
				case 28:
					Key = "bLeft"
					Ignore = GetKeyState(0) & 2^1												// move cursors with alt key
					break
				case 29:
					Key = "bRight"
					Ignore = GetKeyState(0) & 2^1
					break
				case 30:
					Key = "bUp"
					break
				case 31:
					Key = "bDown"
					break
				case 42:	// * key
					PopUpMenu ArithMode	,win=$getPanel() ,mode=3
					break
				case 43:	// + key
					PopUpMenu ArithMode	,win=$getPanel() ,mode=2
					break
				case 45:	// - key
					PopUpMenu ArithMode	,win=$getPanel() ,mode=1
					break
				case 47:	// / key
					PopUpMenu ArithMode	,win=$getPanel() ,mode=4
					break
				case 105:	// i key
					ControlInfo/W=$getPanel() InvertMode
					CheckBox InvertMode	,win=$getPanel() ,value=1-V_Value	
					break
				case 112:	// p key
					Key = "cPin"
					break	
				case 114:	// r key
					Key = "bUndo"
					break
			EndSwitch

			Cursors = CsrInfo(A,getGraph())														// check, if one of the cursors is active: if yes give control back to Igor for cursor movement
			if (strlen(Cursors) && strsearch(Cursors,"/A=0",0,2)  == -1 && Ignore)
				HookTakeover = 0
				break
			endif
			Cursors = CsrInfo(B,getGraph())
			if (strlen(Cursors) && strsearch(Cursors,"/A=0",0,2)  == -1 && Ignore)
				HookTakeover = 0
				break
			endif
			DiffGen_ProcessEvents(Key)															// do the modification according to the pressed key
			break
	EndSwitch
	return HookTakeover
End
//################################################################################################
Function/S DiffGen_LoadWavesAndSettings()														// creates or copies the working waves from the originals
	DFREF stg = root:Packages:$(kDiffGenDir)
	SVar/SDFR=stg locOriginal
	Wave/SDFR=stg Modifiers																		// Modifiers: [0] = horizontal shift, [1] = vertical shift, [2] = multiplicator
	String origname	= ReplaceString("'",ParseFilePath(0, locOriginal, ":", 1, 0),"")
	Wave origwave	= $(locOriginal)
	Wave/Z modwave	= $(ParseFilePath(1, locOriginal, ":", 1, 0) + PossiblyQuoteName(origname + kModEnd))
	
	String recreate = "", outnotes = ""
	if (WaveExists(modwave))																	// duplicate pre-existing mod wave into temp folder
		Duplicate/O modwave stg:WaveO
		recreate = note(modwave)																// extract modify parameters from wave note
	else
		Duplicate/O origwave stg:WaveO
		Modifiers = 0; Modifiers[2] = 1															// reset modify parameters
	endif
	
	Duplicate/O stg:WaveO stg:WaveD, stg:WaveA													// duplicate original to difference and aux waves
	Wave/Z/SDFR=stg WaveO, WaveR, WaveD, WaveA
	WaveD = NaN; WaveA = NaN																	// set to NaNs at first to render the wave invisible
	if (!WaveExists(WaveR))
		Duplicate WaveO stg:WaveR																// create a dummy reference wave
		Wave/SDFR=stg WaveR
		WaveR = NaN
	endif
	
	Variable i
	for (i = 0; i < 4; i += 1)
		Wave/Z outwave	= $(ParseFilePath(1, locOriginal, ":", 1, 0) + PossiblyQuoteName(origname + StringFromList(i,kOutEnd)))
		if (WaveExists(outwave))
			outnotes = note(outwave)
			// *********************** Compatibility code: load old diff waves *******************
			Variable oldDiff = StringMatch(outnotes,"*Difference created from*") + StringMatch(outnotes,"*difference spectrum to*")
			// ***********************************************************************************
			if (!WaveExists(modwave) || oldDiff)												// if no modification wave exits try to load settings from outnote
				recreate = outnotes
			elseif (StringMatch(outnotes,"*"+kRefNameKey+"*"))									// the diff wave was saved with the reference name => switch edit and ref wave
				recreate += "/r"+kRefNameKey
			endif
			break
		endif
	endfor
	// ++++++++++++++++++++++ load saved control settings from wave note +++++++++++++++++++++++++
	Variable origsmth = 0, refsmth = 0, from = 0
	if (strlen(recreate))
		from = strsearch(recreate,"Difference created from:",0)
		from = from == -1 ? 0 : from
		recreate = recreate[from,inf]
		Modifiers[0] = str2num(StringByKey(kShiftKey, recreate, "=", "\r"))
		Modifiers[1] = str2num(StringByKey(kOffsetKey,recreate, "=", "\r"))
		Modifiers[2] = str2num(StringByKey(kScaleKey, recreate, "=", "\r"))
		origsmth = str2num(StringByKey(kModSmthKey,	recreate, "=", "\r"))
		refsmth	 = str2num(StringByKey(kRefSmthKey,	recreate, "=", "\r"))
		origsmth = numtype(origsmth) == 0 ? origsmth : 0
		refsmth  = numtype(refsmth) == 0 ? refsmth : 0
		// ************* Compatibility code: extract mod settings from wave directly *************
		if (numtype(sum(Modifiers)) != 0 && WaveExists(modwave))
			Modifiers[0] = ChangePrecision(DimOffset(modwave,0) - DimOffset(origwave,0),5)
			Modifiers[1] = ChangePrecision(Wavemin(modwave) - Wavemin(origwave),5)
			Modifiers[2] = round(( (Wavemax(modwave) - Wavemin(modwave)) / (Wavemax(origwave) - Wavemin(origwave)))*100000 ) / 100000
		endif
		if (numtype(sum(Modifiers)) != 0)
			Modifiers = 0; Modifiers[2] = 1														// no useful modifiers found
		endif
		// ***************************************************************************************
	endif
	
	String Unit = WaveUnits(origwave,0)
	Variable initY = ChangePrecision(Wavemax(origwave)*0.001,1)									// initial offset value (0.1% of max Y value)
	Variable initX = ChangePrecision(DimDelta(origwave,0),2)									// round to 2 significant digits
	if (strlen(Unit))
		Unit[0] = " "
	endif
	SetVariable Offset		,win=$getPanel() ,value=_NUM:initY		,limits={initY/50,inf,initY/4}	,format="%g"
	SetVariable Shift		,win=$getPanel() ,value=_NUM:initX		,limits={initX/50,inf,initX/4}	,format="%g" + Unit
	SetVariable WaveConv	,win=$getPanel() ,value=_NUM:origsmth	,limits={0,inf,initX/4}			,format="%g" + Unit
	SetVariable RefWaveConv	,win=$getPanel() ,value=_NUM:refsmth	,limits={0,inf,initX/4}			,format="%g" + Unit
	ControlInfo/W=$getPanel() AbsXshift
	if (abs(V_flag) == 5)																		// for newer versions => 4.4
		SetVariable AbsXshift	,win=$getPanel()	,limits={-inf,inf,initX/4}	,format="%g" + Unit
		SetVariable AbsYshift	,win=$getPanel()	,limits={-inf,inf,initY/4}
		SetVariable AbsScale	,win=$getPanel()	,limits={-inf,inf,0.1}
	endif
	CheckBox InvertMode	 	,win=$getPanel() ,value= StringMatch(recreate,"*"+kModeInvKey+"*")
	CheckBox PinBase	 	,win=$getPanel() ,value=!StringMatch(recreate,"*"+kPinBaseKey+"*")
	Variable setMode = 1																		// difference mode is default
	setMode = StringMatch(recreate,"*"+StringFromList(1,kSetModeKey)+"*") ? 2 : setMode
	setMode = StringMatch(recreate,"*"+StringFromList(2,kSetModeKey)+"*") ? 3 : setMode
	setMode = StringMatch(recreate,"*"+StringFromList(3,kSetModeKey)+"*") ? 4 : setMode
	PopUpMenu ArithMode	 	,win=$getPanel() ,mode=setMode
	return recreate																				// copy over to DiffGen_ReloadRefWaveSettings()
End
//################################################################################################
Function DiffGen_ReloadRefWaveSettings(recreate)
	String recreate
	DFREF stg = root:Packages:$(kDiffGenDir)
	SVar/SDFR=stg locOriginal

	Variable switched = StringMatch(recreate,"*"+kRefNameKey+"*")								// the diff wave was saved with the reference name => switch edit and ref wave	
	Variable refsmth  = str2num(StringByKey(kRefSmthKey, recreate, "=", "\r"))
	Variable csrPos	  = str2num(StringByKey(kXCsrSetKey, recreate, "=", "\r")) 
	refsmth = numtype(refsmth) == 0 ? refsmth : 0
	CheckBox AltSaveOption	,win=$getPanel() ,value=switched									// activate ref. name save option in this case
	
	String refWave	= SelectString(switched, StringByKey("Reference Wave", recreate,"=","\r"), StringByKey("Original Wave", recreate,"=","\r"))	// if switched load the original wave into ref. and switch them over later
	String refDir	= ParseFilePath(1, refWave, ":", 1, 0)
	String refName	= ReplaceString("'",ParseFilePath(0, refWave, ":", 1, 0),"")
	String origName	= ReplaceString("'",ParseFilePath(0, locOriginal, ":", 1, 0),"")
	
	// ****************** Compatibility code: extract wave info from old files *******************
	if (StrLen(refDir) == 0 || !DataFolderExists(refDir))
		if (StringMatch(recreate,"*Difference created from*"))
			refDir	= StringByKey("Folder", recreate,":","\r")
			refName	= StringByKey("Wave", recreate,":","\r")
			if (StringMatch(origName,refName))													// don't to load the same wave as both reference and original
				refName = ""
			endif
		elseif(StringMatch(recreate,"*difference spectrum to*"))
			refDir	= GetDataFolder(1)
			refName	= recreate[StrLen("difference spectrum to "),inf]
			refName	= ReplaceString("'",refName, "")
		endif
	endif
	if (StrLen(refDir) == 0 || !DataFolderExists(refDir))										// still not found
		refDir	= GetDataFolder(1)
	endif
	// *******************************************************************************************
	
	DFREF saveDF = GetDataFolderDFR()
	SetDataFolder refDir
	if(Exists(refName) == 1)
		PopUpMenu RefSelect 	,win=$getPanel() ,popmatch=refName								// select the reference wave
		DiffGen_WaveSelect("RefSelect",refName)													// load the old reference wave
		if (switched)
			SwitchWaves()
		endif
		SetVariable RefWaveConv	,win=$getPanel() ,value=_NUM:refsmth							// set ref wave convolution control
		ConvolveRefWave(refsmth)
		if (numtype(csrPos) == 0)
			Cursor/W=$getGraph() A WaveR csrPos
		endif
	endif
	SetDataFolder saveDF
	return 0
End
//################################################################################################
Function DiffGen_UpdateGraphLegends()
	DFREF stg = root:Packages:$(kDiffGenDir)
	Wave/SDFR=stg WaveO, WaveR, WaveD, WaveA
	
	String GraphText = ""
	WaveStats/Q WaveR
	if (V_numNans < DimSize(WaveR,0))															// if nothing is in there, don't generate the legend
		Variable A = -inf,	B = inf
		if (strlen(CsrInfo(A)) > 0 && strlen(CsrInfo(B)) > 0)									// get cursor info if available
			A = xcsr(A)
			B = xcsr(B)
			sprintf GraphText, "Area between %.1f %s and %.1f %s\r", A, WaveUnits(WaveO,0), B, WaveUnits(WaveO,0)
		endif
		MatrixOP/Free tempO = replaceNaNs(WaveO,0) 												// makes sure there are no NaN values
		MatrixOP/Free tempR = replaceNaNs(WaveR,0)
		MatrixOP/Free tempD = replaceNaNs(WaveD,0)
		SetScale/P x, DimOffset(WaveO,0), DimDelta(WaveO,0), tempO
		SetScale/P x, DimOffset(WaveR,0), DimDelta(WaveR,0), tempR
		SetScale/P x, DimOffset(WaveD,0), DimDelta(WaveD,0), tempD
		Variable AreaO = Area(tempO,A,B), AreaR = Area(tempR,A,B), AreaD = Area(tempD,A,B)		// calculate each area
		GraphText += "\\s(WaveO) modified\t(area: "		+ num2str(AreaO)
		GraphText += ")\r\\s(WaveR) reference\t(area: "	+ num2str(AreaR) + "; ratio: " + num2str(abs(AreaR / AreaO))
		GraphText += ")\r\\s(WaveD) difference\t(area: "+ num2str(AreaD) + "; ratio: " + num2str(abs(AreaD / AreaO))
		Switch(kAuxMode)
			case 0:
				GraphText += ")\r\\s(WaveA) smoothed"
				break
			case 1:
				GraphText += ")\r\\s(WaveA) integrated (area: "	+ num2str(Area(WaveA,A,B)) + ")"
				break
		EndSwitch	
	endif
	Legend/C/N=GraphInfoBox/W=$getGraph()/J GraphText											// update trace legend
	DiffGen_UpdateModLegend(getGraph())															// update modification legend
	return 0
End
//################################################################################################
Function DiffGen_UpdateModLegend(Win)
	String Win
	DFREF stg = root:Packages:$(kDiffGenDir)
	Wave/SDFR=stg WaveO, Modifiers																// Modifiers: [0] = horizontal shift, [1] = vertical shift, [2] = multiplicator
	SVar/SDFR=stg locOriginal
	Wave origwave = $locOriginal
	// ++++++++++++++++++++++++ right legend (modification details) ++++++++++++++++++++++++++++++++++++++++++++++
	String BoxText	= "", unit = WaveUnits(WaveO,0)
	if (abs(Modifiers[0]) > 10^-16)
		BoxText += num2str(Modifiers[0]) + " " + SelectString(strlen(unit),"units",unit) + " shift\r"
	endif
	Variable offset = WaveMin(WaveO)-WaveMin(origwave)
	if (abs(offset) > 10^-10)
		BoxText += num2str(offset) + " offset\r"
	endif
	if (Modifiers[2] != 1)
		BoxText += num2str(abs(Modifiers[2])) + " x scaled\r"
	endif
	Variable smth = str2num(StringByKey(kModSmthKey,note(WaveO), "=", "\r"))
	if (numtype(smth) == 0 && smth != 0)
		BoxText += num2str(smth) + SelectString(strlen(unit),""," ") + unit + " smoothed"
	endif
	BoxText[0] = "\K(65280,0,0)"
	Legend/C/N=ModificationInfoBox/W=$Win/J BoxText												// update legend for modification info
	return 0
End
//#################################support functions for control elements#########################
Function DiffGen_PinBaseMode(s) : CheckBoxControl
	STRUCT WMCheckboxAction& s
	if(s.eventCode == 2)
		DiffGen_CalculateDifference(undoPinning=(1-2*s.checked))								// undo the pinning state before doing the calculation
	endif
	return 0
End
//################################################################################################
Function DiffGen_InvertMode(s) : CheckBoxControl
	STRUCT WMCheckboxAction& s
	if(s.eventCode == 2)
		DiffGen_CalculateDifference()
	endif
	return 0
End
//################################################################################################
Function DiffGen_WaveConvolve (s) : SetVariableControl
	STRUCT WMSetVariableAction &s
	if (s.eventCode == 1 || s.eventCode == 2)
		if (StringMatch(s.ctrlName,"RefWaveConv"))												// reference wave is smoothed here, edit wave is smoothed in DiffGen_CalculateDifference()
			ConvolveRefWave(s.dval)
		endif
		DiffGen_CalculateDifference()
	endif
	return 0
End
//################################################################################################
Static Function ConvolveRefWave(value)
	Variable value
	DFREF stg = root:Packages:$(kDiffGenDir)
	SVar/SDFR=stg locReference
	if (strlen(locReference) > 0)
		Wave/SDFR=stg WaveR
		Wave refwave = $(locReference)
		Duplicate/O refwave stg:WaveR
		DiffGen_WaveConvolveGaussHelper(WaveR, value)
	endif
	return 0
End
//############################# convolve with a prepared Gaussian ################################
Function DiffGen_WaveConvolveGaussHelper(in, width)
	Wave in
	Variable width
	if (width == 0 || numtype(width) != 0)
		return 0
	endif
	
	Variable pnt, minVal = WaveMin(in)
	Variable dx  = DimDelta(in,0), size = DimSize(in,0)
	width = abs(width/dx)/sqrt(2)																// width normalized by x scaling and convert to expected Gauss 'width'
	pnt = round(max(abs(10*width),11))															// create a wave 10 times the Gauss width to get the full Gauss in (5 sigma on each side)
	pnt = min(pnt, 2*numpnts(in)+1)																// doesn't have to be larger than two times the full input data
	pnt = pnt+!mod(pnt, 2)																		// force odd size

	Make/FREE/D/N=(pnt) GaussWave	
	GaussWave = Gauss(x, (pnt-1)/2, width)
	Variable A = sum(GaussWave)																	// make sure the Gauss area limited even for a few points
	if (A > 1)
		GaussWave /= A
	endif
	
	Variable NaNs = 0, num = NaN																// find NaNs at the start and end points
	do
		num = in[NaNs]
		NaNs += 1
	while (numtype(num) != 0 && NaNs < size-1)
	do
		num = in[size-1]
		size -= 1
	while (numtype(num) != 0 && size > 0)
	NaNs -= 1
	size += 1
	
	Make/FREE/D/N=(2*pnt+size-NaNs) temp = in[NaNs]												// enlarge the wave to get the full convolution range
	temp[pnt,size+pnt-1-NaNs] = in[p-pnt+NaNs]
	temp[size+pnt-NaNs,size+2*pnt-1-NaNs] = in[size-1]
	MatrixOP/Free temp = replaceNaNs(temp,0)													// makes sure there are no NaN values in-between
	
	Convolve/A GaussWave temp
	in[NaNs,size-1] = temp[p+pnt-NaNs]
	
	return WaveMin(in)-minVal																	// return by how much the baseline has shifted
End
//################################################################################################
Function DiffGen_ModeSelect(s) : PopupMenuControl
	STRUCT WMPopupAction &s
	if (s.eventCode == 2)
		DiffGen_CalculateDifference()
		ModifyControl bUndo ,win=$getPanel() ,activate											// move focus away from the popup (prevents accidental selection of a different wave with the cursor keys)
	endif
	return 0
End
//################################################################################################
Function DiffGen_PopupWaveSelect(s) : PopupMenuControl
	STRUCT WMPopupAction &s
	if (s.eventCode == 2 && (s.popNum > 1 || StringMatch(s.popStr,"none")))
		DiffGen_WaveSelect(s.ctrlName, s.popStr)
		ModifyControl bUndo ,win=$getPanel() ,activate											// move focus away from the popup (prevents accidental selection of a different wave with the cursor keys)
	endif
	return 0
End
//+++++++++++++++++++++++++++++++ popup help function +++++++++++++++++++++++++++++++++++++
Function/S DiffGen_CreateWaveList(Type)															// help function to create the wave lists for the drop down menus
	String Type
	SVar locOriginal = root:Packages:$(kDiffGenDir):locOriginal
	String name	 = ReplaceString("'",ParseFilePath(0, locOriginal, ":", 1, 0),"")
	String waves = WaveList("!" + name,";", "DIMS:1,TEXT:0")									// exclude original wave name (prevents creating difference with identical wave)
	waves = SortList(waves, ";", 16)
	ControlInfo/W=$getPanel() ArithMode
	waves = RemoveFromList(name+StringFromList(V_value-1,kOutEnd)+";"+name+kModEnd+";"+name+StringFromList(kAuxMode,kAuxEnd)+";",waves)	// remove the associated result waves (cannot remove reference name here or switching will not work)
	return SelectString(StringMatch(Type,"Ref"), name + ";", "none;") + waves					// in case of reference wave popup - create wave list from current folder without original wave
End
//################################################################################################
Function DiffGen_WaveSelect(type, name)
	String type, name
	
	DFREF stg = root:Packages:$(kDiffGenDir)
	SVar/SDFR=stg locOriginal, locReference
	
	StrSwitch (type)
		case "OrigSelect":
			locOriginal = GetWavesDataFolder($name, 2)
			DiffGen_LoadWavesAndSettings()														// copy waves into working folder
			DoWindow /T $getGraph(), kDiffTitle + name											// change the window title
		break
		case "RefSelect":
			Wave/SDFR=stg WaveR
			locReference = ""; WaveR = NaN														// reset previous reference
			RemoveFromGraph/W=$getGraph()/Z WaveR
			if (!StringMatch(name,"none"))
				locReference = GetWavesDataFolder($name, 2)
				Duplicate/O $name stg:WaveR
				AppendToGraph/W=$getGraph()/C=(0,0,65280) WaveR
				
				ControlInfo/W=$getPanel() RefWaveConv
				if (V_Value > 0)
					ConvolveRefWave(V_Value)
				endif
			endif
		break
	endswitch
	DiffGen_CalculateDifference()
	return 0
End
//################################################################################################
Static Function SwitchWaves()																	// switches reference and edit wave selection
	DFREF stg = root:Packages:$(kDiffGenDir)
	SVar/SDFR=stg locOriginal, locReference
	Wave OrigWave  = $locOriginal
	Wave/Z RefWave = $locReference

	ControlInfo/W=$getPanel() OrigSelect
	String OrigSelection = S_Value	
	ControlInfo/W=$getPanel() RefSelect
	String RefSelection = S_Value

	DFREF dfSav = GetDataFolderDFR()	
	if (!stringmatch(RefSelection,"none"))														// switch waves by selecting the reference in the original's place and vice versa
		SetDataFolder GetWavesDataFolderDFR(RefWave)
		PopUpMenu OrigSelect ,win=$getPanel() ,popMatch=RefSelection
		DiffGen_WaveSelect ("OrigSelect",RefSelection)
		
		SetDataFolder GetWavesDataFolderDFR(OrigWave)
		PopUpMenu RefSelect	 ,win=$getPanel() ,popMatch=OrigSelection
		DiffGen_WaveSelect ("RefSelect",OrigSelection)
	endif
	SetDataFolder dfSav
	return 0
End
//################################################################################################
Function DiffGen_ButtonControl(s) : ButtonControl
	STRUCT WMButtonAction &s
	if (s.eventCode == 2)
		if (StringMatch(s.ctrlName,"bProg*"))
			DiffGen_ProgramExecute(s.ctrlName)													// these buttons have a different execute function (no difference calculation)
		else
			DiffGen_ProcessEvents(s.ctrlName)
		endif
	endif
	return 0
End
//+++++++++++++++++++++++++++++++++++++ button help function +++++++++++++++++++++++++++++++++++++
Function DiffGen_ProgramExecute(type)
	String type
	StrSwitch(type)
		case "bProgSaveQuit":
			if (WinType(getGraph()+"#AskQuitPanel") == 7)
				KillWindow $(getGraph()+"#AskQuitPanel")
			endif
			DiffGen_SaveAndGraphResult()														// no break here => also kill window										
			if (GetKeyState(0) == 1)															// do not quit when ctrl is pressed
				break
			endif
		case "bProgUndoQuit":
			SetWindow $getPanel() hook(DiffWindow)=$""											// release hook to not trigger any events
			SetWindow $getGraph() hook(DiffWindow)=$""
			KillWindow $getGraph()																// really kill the window, before deleting everything
			DFREF stg = root:Packages:$(kDiffGenDir)
			DFREF pkg = root:Packages:
			KillDataFolder/Z stg
			Variable pkgObjects = CountObjectsDFR(pkg, 1)+CountObjectsDFR(pkg, 2)+CountObjectsDFR(pkg, 3)+CountObjectsDFR(pkg, 4)
			if (pkgObjects == 0)																// clean up Packages folder if empty
				KillDataFolder pkg
			endif
			break
		case "bProgCancel":																		// just remove the sub-panel
			KillWindow #AskQuitPanel
			break
		case "bProgSwitch":																		// the switch button
			SwitchWaves()
			break
	EndSwitch
	return 0
End
//+++++++++++++++++++++++++++++++++++++ button help function +++++++++++++++++++++++++++++++++++++
Function DiffGen_ProcessEvents(ctrlName)														// update modification parameters and recalculate difference
	String ctrlName
	
	DFREF stg = root:Packages:$(kDiffGenDir)
	Wave/SDFR=stg Modifiers																		// Modifiers: [0] = horizontal shift, [1] = vertical shift, [2] = multiplicator

	Variable MinScaleLimit = 0.0001																// lower limit for scaling (multiplication)
	Variable Accelerate = 1, Rev = 1															// acceleration modifier & reverse action modifier (to trigger the opposite effect)
	Variable HoldKey = GetKeyState(0)
	Accelerate = (HoldKey & 2^2) != 0 ? 10 : Accelerate											// when additional keys are pressed, modify the experience
	Accelerate = (HoldKey & 2^0) != 0 ? 0.5 : Accelerate										// gets faster (shift = 10x) or slower (ctrl = 1/2)
	Accelerate = ((HoldKey & 2^0) && (HoldKey & 2^2)) != 0 ? 0.1 : Accelerate					// super slow (1/10) with both shift and ctrl
	
	Strswitch (ctrlName)																		// the action for each button 
		case "bDown":
			Rev = -1
		case "bUp":
			ControlInfo/W=$getPanel() Offset
			Modifiers[1] += V_Value * Accelerate * Rev
			Modifiers[1] = abs(Modifiers[1]) < 10^-16 ? 0 : Modifiers[1]						// get rid of floating point errors
		break
		case "bLeft":
			Rev = -1
		case "bRight":
			ControlInfo/W=$getPanel() Shift
			Modifiers[0] += V_Value * Accelerate * Rev
			Modifiers[0] = abs(Modifiers[0]) < 10^-16 ? 0 : Modifiers[0]
		break
		case "bStretch":
			Rev = -1
		case "bShrink":
			ControlInfo/W=$getPanel() Multiplier
			Modifiers[2]  = (Modifiers[2] == MinScaleLimit && Rev == -1) ? 0 : Modifiers[2]		// scroll away smoothly from minimum limit
			Modifiers[2] -= (V_Value * Accelerate * Rev)/100									// do the modification
			Modifiers[2]  = Modifiers[2] < MinScaleLimit ? MinScaleLimit : Modifiers[2]			// make sure not to shrink the data to zero or invert
		break
		case "bUndo":																			// resets the modifications by overwriting with original data
			Modifiers = 0; Modifiers[2] = 1														// reset modify parameters
			SetVariable WaveConv	,win=$getPanel() ,value=_NUM:0								// reset convolution control
			SetVariable RefWaveConv	,win=$getPanel() ,value=_NUM:0
			if((GetKeyState(0) & 2^0) != 0)														// also reset reference wave to none if ctrl key is pressed
				PopUpMenu RefSelect ,win=$getPanel() ,mode=1
				DiffGen_WaveSelect ("RefSelect","none")
			else
				ConvolveRefWave(0)
			endif
		break
		case "bUndoShift":
			Modifiers[0] = 0
		break
		case "bUndoOffset":
			Modifiers[1] = 0
		break
		case "bUndoScale":
			Modifiers[2] = 1
		break
		case "cPin":
			ControlInfo/W=$getPanel() PinBase
			CheckBox PinBase ,win=$getPanel() ,value=1-V_Value
			Variable minPoint, minCsrX; DiffGen_FindBaseline(minPoint, minCsrX)
			Rev = V_Value ? 1 : -1
			Modifiers[1] += rev*minPoint*(1 - Modifiers[2])										// add or subtract the scaled amount to the offset (prevents 'snapping')
		break
	Endswitch
	
	DiffGen_CalculateDifference()
	return 0
End
//################################################################################################
Function DiffGen_SaveAndGraphResult()
	DFREF stg = root:Packages:$(kDiffGenDir)
	SVar/SDFR=stg locOriginal, locReference
	Wave/SDFR=stg/Z WaveO, WaveD, WaveA	
	ControlInfo/W=$getPanel() ArithMode
	Variable setMode = V_Value - 1
	String savedir	= ParseFilePath(1, locOriginal, ":", 1, 0)
	String origname	= ReplaceString("'",ParseFilePath(0, locOriginal, ":", 1, 0),"")
	String modname	= PossiblyQuoteName(origname + kModEnd)
	String diffname = PossiblyQuoteName(origname + StringFromList(setMode,kOutEnd))
	String auxname	= PossiblyQuoteName(origname + StringFromList(kAuxMode,kAuxEnd))
	Wave origwave	= $(locOriginal)
	Wave/Z refwave	= $(locReference)
	Wave/Z modwave	= $(savedir + modname)
	Wave/Z diffwave	= $(savedir + diffname)
	Wave/Z auxwave	= $(savedir + auxname)
	
	KillWaves/Z modwave, diffwave, auxwave														// try to remove previous versions (cleans up the folder)
	if (!EqualWaves(WaveO, origwave, 5))														// copy the modified wave only when something has changed
		Duplicate/O WaveO $(savedir + modname)
		Wave origwave = $(savedir + modname)													// for display purposes
		origname = modname																		// for legend update
	endif
	
	WaveStats/Q WaveD
	if (V_numNans != DimSize(WaveD,0))															// only if there is more than NaNs, copy the difference
		ControlInfo/W=$getPanel() AltSaveOption
		if (GetKeyState(0) & 2^1 || V_Value)													// keyboard shortcut: Press alt to write the difference wave with the reference wave as name
			diffname = PossiblyQuoteName(NameOfWave(refwave) + StringFromList(setMode,kOutEnd))
			auxname	 = PossiblyQuoteName(NameOfWave(refwave) + StringFromList(kAuxMode,kAuxEnd))
			savedir	 = ParseFilePath(1, locReference, ":", 1, 0)								// change to ref folder
			Note WaveD, "\rReference name used"
		endif
		Duplicate/O WaveD $(savedir + diffname)
		Wave diffwave = $(savedir + diffname)
		if (kSaveAux)
			Duplicate/O WaveA $(savedir + auxname)
			Wave auxwave = $(savedir + auxname)
		endif
	else
		return 0
	endif
	// ++++++++++++++++++++++++++ create and update summary graphs +++++++++++++++++++++++++++++++
	ControlInfo/W=$getPanel() FinishGraph
	if (V_Value)
		Display/K=1/W=(200,50,700,300) origwave as "Result of " + NameOfWave(origwave)			// display the results
		AppendToGraph/C=(0,0,65280) refwave
		AppendToGraph/C=(0,0,0) diffwave
		if (kSaveAux)
			AppendToGraph/C=(0,40000,20000) auxwave
		endif
		ModifyGraph mirror(left)=1, mirror(bottom)=2, minor=1, standoff=0, ZisZ=1, notation(left)=1	
		Legend/C/N=ModificationInfoBox/F=0/A=RT/X=2.00/Y=2.00/J/B=1								// move legend to the right
		Legend/C/N=GraphInfoBox/F=0/A=LT/X=2.00/Y=2.00											// generate default legend
	endif
	
	String list = WinList("*", ";","WIN:1")
	Variable win
	for (win = 0; win < ItemsInList(list); win += 1)											// update the modification labels of all graphs with this data set
		String currWin = StringFromList(win, list)
		Variable matchWave = Stringmatch(TraceNameList(currWin,";",1),"*"+origname+"*")
		Variable matchAnno = Stringmatch(AnnotationList(currWin),"*ModificationInfoBox*")
		if (matchWave && matchAnno)
			DiffGen_UpdateModLegend(currWin)
		endif
	endfor
	return 0
End